const GAME = { };
let GAME_DATA = null;

GAME.Board = class {
    constructor() {
        this.tiles = [ ]; for(let i = 0; i < 61; ++i) { this.tiles.push(new GAME.Tile(i)); }
        this.pieces = [ ]; for(let i = 0; i < 38; ++i) { this.pieces.push(null); }
        this.columns = [ ]; for(let i = 0; i < 9; ++i) { this.columns.push(new GAME.Column()); }

        this.init();
    }

    init() {
        this.pieces = [ ]; for(let i = 0; i < 38; ++i) { this.pieces.push(null); }

        // Describe Dawn layout
        let layout = [
            { piece:GAME.Const.PieceId.Militia, hex:new MATH.Vec2(0, 1) },
            { piece:GAME.Const.PieceId.Militia, hex:new MATH.Vec2(1, 1) },
            { piece:GAME.Const.PieceId.Militia, hex:new MATH.Vec2(2, 2) },
            { piece:GAME.Const.PieceId.Militia, hex:new MATH.Vec2(3, 2) },
            { piece:GAME.Const.PieceId.Militia, hex:new MATH.Vec2(4, 3) },
            { piece:GAME.Const.PieceId.Militia, hex:new MATH.Vec2(5, 3) },
            { piece:GAME.Const.PieceId.Militia, hex:new MATH.Vec2(6, 4) },
            { piece:GAME.Const.PieceId.Militia, hex:new MATH.Vec2(7, 4) },
            { piece:GAME.Const.PieceId.Militia, hex:new MATH.Vec2(8, 5) },

            { piece:GAME.Const.PieceId.Lance, hex:new MATH.Vec2(0, 0) },
            { piece:GAME.Const.PieceId.Lance, hex:new MATH.Vec2(8, 4) },

            { piece:GAME.Const.PieceId.Knight, hex:new MATH.Vec2(1, 0) },
            { piece:GAME.Const.PieceId.Knight, hex:new MATH.Vec2(7, 3) },

            { piece:GAME.Const.PieceId.Castle, hex:new MATH.Vec2(2, 0) },
            { piece:GAME.Const.PieceId.Castle, hex:new MATH.Vec2(6, 2) },

            { piece:GAME.Const.PieceId.Tower, hex:new MATH.Vec2(3, 0) },
            { piece:GAME.Const.PieceId.Tower, hex:new MATH.Vec2(5, 1) },

            { piece:GAME.Const.PieceId.Dragon, hex:new MATH.Vec2(4, 1) },
            { piece:GAME.Const.PieceId.Omen, hex:new MATH.Vec2(4, 0) },
        ];

        // Add Dawn pieces
        for(let piece of layout) {
            this.set_piece(piece.piece, GAME.Const.Player.Dawn, piece.hex);
        }

        // Add Dusk pieces
        for(let piece of layout) {
            let hex = new MATH.Vec2(8 - piece.hex.x, 8 - piece.hex.y);
            this.set_piece(piece.piece, GAME.Const.Player.Dusk, hex);
        }
    }

    set_piece(piece, player, hex) {
        let index = 0;
        while(this.pieces[index] !== null) { index++; }

        let game_piece = new GAME.Piece(piece, player);
        game_piece.tile = HEX.hex_to_tile(hex);
        this.tiles[game_piece.tile].piece = index;
        this.pieces[index] = game_piece;
    }

    reset() {
        for(let i = 0; i < this.tiles.length; ++i) { this.tiles[i].reset(); }
        for(let i = 0; i < this.columns.length; ++i) { this.columns[i].reset(); }
    }
};

GAME.Pool = class {
    constructor() {
        this.pieces = [ ]; for(let i = 0; i < 6; ++i) { this.pieces.push(1); }
    }
};

GAME.Column = class {
    constructor() {
        this.militia = [false, false];
        this.extent = [0, 8];
    }

    reset() {
        this.militia = [false, false];
        this.extent = [0, 8];
    }
};

GAME.Tile = class {
    constructor(index) {
        this.piece = null;

        this.threaten = [false, false];
        this.checking = [false, false];

        this.hex = HEX.tile_to_hex(index);
    }

    reset() {
        this.threaten = [0, 0];
        this.checking = [false, false];
    }
};

GAME.MovementTile = class {
    constructor(tile, status) {
        this.tile = tile;
        this.status = status;
    }
};

GAME.Play = class {
    constructor(source, from, to) {
        this.source = source;
        this.from = from;
        this.to = to;
    }
};

GAME.GamePiece = class {
    constructor(name, moves, promote_moves) {
        this.name = name;
        this.moves = moves;
        this.pmoves = promote_moves;
    }
};

GAME.PieceMovement = class {
    constructor() {
        this.direction = 0;
        this.stride = 0;
    }

    add(direction) {
        this.direction |= 1 << direction;
        return this;
    }

    add_stride(direction) {
        this.direction |= 1 << direction;
        this.stride |= 1 << direction;
        return this;
    }

    rotate() {
        let copy = new GAME.PieceMovement();
        copy.direction = BITWISE.rotate_blocks(this.direction);
        copy.stride = BITWISE.rotate_blocks(this.stride);
        return copy;
    }
};

GAME.Piece = class {
    constructor(piece, player) {
        this.piece = piece;
        this.player = player;
        this.promoted = false;

        this.tile = 0;
        this.blocking = 0;
    }

    moves() {
        let moves = null;
        if(this.promoted) { moves = GAME.Const.Piece[this.piece].pmoves; }
        else { moves = GAME.Const.Piece[this.piece].moves; }
        if(this.player == GAME.Const.Player.Dusk) { moves = moves.rotate(); }
        return moves;
    }

    reset() {
        this.blocking = 0;
    }
};

GAME.Game = class {
    constructor() {
        this.turn = 0;
        
        this.board = new GAME.Board();
        this.pools = [
            new GAME.Pool(),
            new GAME.Pool(),
        ];

        this.state = {
            check:false,
            checkmate:false,
        };

        this.update_board();
    }

    update_board() {
        // Reset tiles
        this.board.reset();

        this.state.check = false;
        this.state.checkmate = false;

        // Determine threaten, check, and blocking for each piece
        for(let piece of this.board.pieces) {
            if(piece !== null) {
                let hex = this.board.tiles[piece.tile].hex;

                // Check if column has militia.
                if(piece.piece == GAME.Const.PieceId.Militia) {
                    this.board.columns[hex.x].militia[piece.player] = true;
                }

                // Check furthest piece in column.
                if(piece.player == 0) { this.board.columns[hex.x].extent[0] = Math.max(hex.y, this.board.columns[hex.x].extent[0]); }
                else { this.board.columns[hex.x].extent[1] = Math.min(hex.y, this.board.columns[hex.x].extent[1]); }

                for(let movement of this.movement_tiles_ext(piece, piece.tile)) {
                    if(movement.status != GAME.Const.MoveStatus.Invalid) {
                        if(movement.status == GAME.Const.MoveStatus.Check) {
                            this.state.check = true;
                        }
                        this.board.tiles[movement.tile].threaten[piece.player] = true;
                    }
                }
            }
        }

        // Count moves available to next turn player to determine checkmate.
        if(this.state.check) {

        }
    }

    process(move) {
        // Check if swapped piece.
        // Check if piece should be promoted.
        // Check if swapped piece should be promoted.
        // Add piece to pool if taken.
        // Move pieces.

        // TODO
        //  - Validate move.
        //  - Improve data safety validation.
        //

        // Move piece on board.
        if(move.source == 0) {
            let piece_id = this.board.tiles[move.from].piece;
            let piece = this.board.pieces[piece_id];
            piece.tile = move.to;
            this.board.tiles[move.from].piece = null;

            let target_id = this.board.tiles[move.to].piece;
            if(target_id !== null) {
                let target = this.board.pieces[target_id];
                
                // Swap piece with moving piece.
                if(target.player == piece.player) {
                    target.tile = move.from;
                    this.board.tiles[move.from].piece = target_id;

                    // Check if swap is promoted.
                    let hex = HEX.tile_to_hex(target.tile);
                    hex.y -= MATH.sign_branch(target.player);
                    if(!target.promoted && !HEX.is_valid(hex)) {
                        target.promoted = true;
                    }
                }
                
                // Add captured piece to pool and destroy.
                else {
                    console.log(piece.player);
                    this.pools[piece.player].pieces[target.piece] += 1;
                    this.board.pieces[target_id] = null;
                }
            }

            this.board.tiles[move.to].piece = piece_id;

            // Check if piece is promoted.
            let hex = HEX.tile_to_hex(piece.tile);
            hex.y -= MATH.sign_branch(piece.player);
            if(!piece.promoted && !HEX.is_valid(hex)) {
                piece.promoted = true;
            }
        }

        // Place piece from pool.
        else {

        }

        this.turn++;

        // Recalculate new board state.
        this.update_board();

        // If check, detect checkmate.
    }

    movement_tiles(piece, tile) {
        let tiles = movement_tiles_ext(piece, tile);
        let result = [ ];

        for(let movement in tiles) {
            if(movement.status) { result.push(movement); }
        }

        return result;
    }

    movement_tiles_ext(piece, tile) {
        let tiles = [ ];
        let moves = piece.moves();
        let hex = this.board.tiles[tile].hex;

        let directions = moves.direction;
        let permitted_moves =  directions;
        if(piece.blocking != 0) {
            if(piece.piece == GAME.Const.PieceId.Omen) {
                permitted_moves &= ~piece.blocking;
            } else {
                permitted_moves &= piece.blocking;
            }
        }

        // Check directions of movement.
        for(let mask = BITWISE.lsb(directions); directions > 0; mask = BITWISE.lsb(directions)) {
            let direction_id = BITWISE.ffs(mask);
            let direction = GAME.Const.get_direction(direction_id);
            let stride = (moves.stride & mask) != 0;

            // Get initial status in direction.
            let status = (permitted_moves & mask)? GAME.Const.MoveStatus.Valid : GAME.Const.MoveStatus.Invalid;

            let move_hex = hex.copy();

            // Check tiles in direction up to movement limit.
            let max_dist = (stride)? 8 : 1;
            for(let dist = 1; dist <= max_dist; ++dist) {
                move_hex.add(direction);

                if(HEX.is_valid(move_hex)) {
                    let tile_id = HEX.hex_to_tile(move_hex);
                    let tile_data = this.board.tiles[tile_id];

                    let result = status;

                    // King may not move onto threatened tile.
                    if(piece.piece == GAME.Const.PieceId.Omen && tile_data.threaten[+(!piece.player)]) {
                        status = GAME.Const.MoveStatus.Invalid;
                    }

                    // Handle occupied tile.
                    if(tile_data.piece !== null) {
                        let target = this.board.pieces[tile_data.piece];

                        // Target piece is ally.
                        if(target.player == piece.player) {

                            // Move is only valid if pieces are swappable.
                            if(!this.movement_swappable(target, mask, dist, tile)) {
                                result = GAME.Const.MoveStatus.Invalid;
                            }
                            status = GAME.Const.MoveStatus.Invalid;
                        }
                        
                        // Target piece is opposing.
                        else {
                            // Check if target piece is opposing king.
                            if(target.piece == GAME.Const.PieceId.Omen) {
                                if(status == GAME.Const.MoveStatus.Valid) {
                                    result = GAME.Const.MoveStatus.Check;
                                }
                            }
                            status = GAME.Const.MoveStatus.Invalid;
                        }
                    }

                    tiles.push(new GAME.MovementTile(tile_id, result));
                } else { break; }
            }

            directions &= ~mask;
        }

        return tiles;
    }

    movement_swappable(target, mask, range, tile) {
        mask = BITWISE.rotate_blocks(mask);
        let moves = target.moves();

        // King cannot swap onto tile that is threatened.
        // This case should also cover blocking.
        if(target.piece == GAME.Const.PieceId.Omen && this.board.tiles[tile].threaten[+(!target.player)]) {
            return false;
        }
        return ((moves.direction & mask) != 0 && (range == 1 || (moves.stride & mask) != 0));
    }

    placement_tiles_ext(piece_id, player) {
        let tiles = [ ];
        let piece = new GAME.Piece(piece_id, player);

        // Get tiles onto which piece may be placed.
        for(let i = 0; i < this.board.tiles.length; ++i) {
            let hex = HEX.tile_to_hex(i);
            let tile = this.board.tiles[i];
            let status = GAME.Const.MoveStatus.Invalid;

            // Check if tile is occupied.
            if(tile.piece === null) {
                let position_valid = true;
                
                // Check off-sides.
                if(piece.player == 0) {
                    position_valid = hex.y <= this.board.columns[hex.x].extent[+(!player)];
                } else {
                    position_valid = hex.y >= this.board.columns[hex.x].extent[+(!player)];
                }

                // Check militia stacking.
                if(piece_id == GAME.Const.PieceId.Militia && this.board.columns[hex.x].militia[player]) {
                    position_valid = false;
                }

                // Check if position puts king in check.
                let checking = false;
                let movements = this.movement_tiles_ext(piece, i);
                for(let movement of movements) {
                    if(movement.status == GAME.Const.MoveStatus.Check) {
                        checking = true;
                        break;
                    }
                }

                // Piece must have movements and not put king in check.
                if(position_valid && movements.length > 0 && !checking) {
                    status = GAME.Const.MoveStatus.Valid;
                }
            }

            tiles.push(new GAME.MovementTile(i, status));
        }

        return tiles;
    }
};

GAME.Const = {
    Player: {
        Dawn: 0,
        Dusk: 1,
    },

    Source: {
        Board: 0,
        Pool: 1,
    },

    Direction: [
        new MATH.Vec2(0, 1),
        new MATH.Vec2(1, 1),
        new MATH.Vec2(1, 0),
        new MATH.Vec2(0, -1),
        new MATH.Vec2(-1, -1),
        new MATH.Vec2(-1, 0),

        new MATH.Vec2(1, 2),
        new MATH.Vec2(2, 1),
        new MATH.Vec2(1, -1),
        new MATH.Vec2(-1, -2),
        new MATH.Vec2(-2, -1),
        new MATH.Vec2(-1, 1),
    ],

    MoveStatus: {
        Valid: 0,
        Invalid: 1,
        Check: 2,
    },

    PieceId: {
        Militia: 0,
        Knight: 1,
        Lance: 2,
        Tower: 3,
        Castle: 4,
        Dragon: 5,
        Omen: 6,
    },

    Piece: [
        new GAME.GamePiece(
            "Militia",
            new GAME.PieceMovement()
                .add(0)
                .add(1)
                .add(5),
            new GAME.PieceMovement()
                .add(0)
                .add(1)
                .add(2)
                .add(4)
                .add(5),
        ),
        new GAME.GamePiece(
            "Knight",
            new GAME.PieceMovement()
                .add(3)
                .add(6)
                .add(11)
                .add(13)
                .add(17),
            new GAME.PieceMovement()
                .add(3)
                .add(6)
                .add(7)
                .add(10)
                .add(11)
                .add(13)
                .add(14)
                .add(16)
                .add(17),
        ),
        new GAME.GamePiece(
            "Lance",
            new GAME.PieceMovement()
                .add_stride(0)
                .add(1)
                .add(5),
            new GAME.PieceMovement()
                .add_stride(0)
                .add_stride(1)
                .add_stride(2)
                .add_stride(3)
                .add_stride(4)
                .add_stride(5),
        ),
        new GAME.GamePiece(
            "Tower",
            new GAME.PieceMovement()
                .add(0)
                .add(1)
                .add(3)
                .add(5)
                .add(6)
                .add(11),
            new GAME.PieceMovement()
                .add(0)
                .add(1)
                .add(2)
                .add(3)
                .add(4)
                .add(5)
                .add(6)
                .add(8)
                .add(9)
                .add(11),
        ),
        new GAME.GamePiece(
            "Castle",
            new GAME.PieceMovement()
                .add(0)
                .add(1)
                .add(2)
                .add(4)
                .add(5)
                .add(7)
                .add(10),
            new GAME.PieceMovement()
                .add(0)
                .add(1)
                .add(2)
                .add(3)
                .add(4)
                .add(5)
                .add_stride(7)
                .add_stride(10),
        ),
        new GAME.GamePiece(
            "Dragon",
            new GAME.PieceMovement()
                .add_stride(6)
                .add_stride(7)
                .add_stride(8)
                .add_stride(9)
                .add_stride(10)
                .add_stride(11),
            new GAME.PieceMovement()
                .add_stride(0)
                .add_stride(1)
                .add_stride(2)
                .add_stride(3)
                .add_stride(4)
                .add_stride(5)
                .add_stride(6)
                .add_stride(7)
                .add_stride(8)
                .add_stride(9)
                .add_stride(10)
                .add_stride(11),
        ),
        new GAME.GamePiece(
            "Omen",
            new GAME.PieceMovement()
                .add(0)
                .add(1)
                .add(2)
                .add(3)
                .add(4)
                .add(5)
                .add(7)
                .add(10),
        ),
    ],

    get_direction(direction_id) {
        let direction = GAME.Const.Direction[direction_id % 12].copy();
        direction.mul(Math.ceil((direction_id + 1) / 12));
        return direction;
    },
};

GAME.init = () => {
    GAME_DATA = new GAME.Game();
};
